@@ -8,6 +8,7 @@ from typing import cast, TYPE_CHECKING
 
 from nbsafety.data_model.annotation_utils import get_type_annotation, make_annotation_string
 from nbsafety.data_model import sizing
+from nbsafety.data_model.timestamp import Timestamp
 from nbsafety.data_model.update_protocol import UpdateProtocol
 from nbsafety.singletons import nbs, tracer
 
@@ -35,6 +36,16 @@ class DataSymbol:
 
     NULL = object()
 
+    IMMUTABLE_TYPES = {
+        bytes,
+        bytearray,
+        float,
+        frozenset,
+        int,
+        str,
+        tuple,
+    }
+
     def __init__(
         self,
         name: SupportedIndexType,
@@ -72,22 +83,20 @@ class DataSymbol:
         # initialize at -1 since the corresponding piece of data could already be around,
         # and we don't want liveness checker to think this was newly created unless we
         # explicitly trace an update somewhere
-        self._timestamp: int = -1
+        self._timestamp: Timestamp = Timestamp.uninitialized()
         # The version is a simple counter not associated with cells that is bumped whenever the timestamp is updated
         self._version: int = 0
-        self._defined_cell_num = self._timestamp
+        self._defined_cell_num = nbs().cell_counter()
 
         # The necessary last-updated timestamp / cell counter for this symbol to not be stale
-        self.required_timestamp: int = self.timestamp
+        self.required_timestamp: Timestamp = self.timestamp
 
-        # The execution counter of cell where this symbol was last used (-1 means it has net yet been used)
-        self.last_used_cell_num: int = -1
         # for each usage of this dsym, the version that was used, if different from the timestamp of usage
-        self.timestamp_by_used_time: Dict[int, int] = {}
+        self.timestamp_by_used_time: Dict[Timestamp, Timestamp] = {}
         # History of definitions at time of liveness
-        self.timestamp_by_liveness_time: Dict[int, int] = {}
+        self.timestamp_by_liveness_time_by_cell_counter: Dict[int, Dict[Timestamp, Timestamp]] = defaultdict(dict)
         # All timestamps associated with this symbol
-        self.updated_timestamps: Set[int] = set()
+        self.updated_timestamps: Set[Timestamp] = set()
 
         self.fresher_ancestors: Set[DataSymbol] = set()
 
@@ -119,6 +128,13 @@ class DataSymbol:
     def temporary_disable_warnings(self):
         self._temp_disable_warnings = True
 
+    @property
+    def last_used_timestamp(self):
+        if len(self.timestamp_by_used_time) == 0:
+            return Timestamp.uninitialized()
+        else:
+            return max(self.timestamp_by_used_time.keys())
+
     @property
     def namespace_stale_symbols(self) -> Set[DataSymbol]:
         ns = self.namespace
@@ -129,11 +145,15 @@ class DataSymbol:
         return self._timestamp
 
     @property
-    def timestamp(self) -> int:
+    def timestamp(self) -> Timestamp:
         ts = self._timestamp
         ns = self.namespace
         return ts if ns is None else max(ts, ns.max_descendent_timestamp)
 
+    @property
+    def defined_cell_num(self) -> int:
+        return self._defined_cell_num
+
     @property
     def readable_name(self) -> str:
         return self.containing_scope.make_namespace_qualified_name(self)
@@ -308,6 +328,11 @@ class DataSymbol:
         if refresh_cached:
             self._refresh_cached_obj()
 
+    def invalidate_cached(self):
+        self._cached_out_of_sync = True
+        self.cached_obj_id = None
+        self.cached_obj_type = None
+
     def get_ref_count(self):
         if self.obj is None:
             return -1
@@ -370,7 +395,7 @@ class DataSymbol:
     def _match_call_args_with_definition_args(self):
         assert self.is_function and isinstance(self.stmt_node, (ast.FunctionDef, ast.AsyncFunctionDef, ast.Lambda))
         caller_node = self._get_calling_ast_node()
-        if not isinstance(caller_node, ast.Call):
+        if caller_node is None or not isinstance(caller_node, ast.Call):
             return []
         def_args = self.stmt_node.args.args
         if len(self.stmt_node.args.defaults) > 0:
@@ -398,7 +423,7 @@ class DataSymbol:
 
     def _get_calling_ast_node(self) -> Optional[ast.AST]:
         if isinstance(self.stmt_node, (ast.FunctionDef, ast.AsyncFunctionDef)):
-            if self.name == '__getitem__':
+            if self.name in ('__getitem__', '__setitem__', '__delitem__'):
                 # TODO: handle case where we're looking for a subscript for the calling node
                 return None
             for decorator in self.stmt_node.decorator_list:
@@ -446,19 +471,21 @@ class DataSymbol:
     def update_deps(
         self,
         new_deps: Set[DataSymbol],
-        prev_obj=None,
-        overwrite=True,
-        mutated=False,
-        deleted=False,
-        propagate_to_namespace_descendents=False,
-        propagate=True,
-        refresh=True,
-    ):
+        prev_obj: Any = None,
+        overwrite: bool = True,
+        mutated: bool = False,
+        deleted: bool = False,
+        propagate_to_namespace_descendents: bool = False,
+        propagate: bool = True,
+        refresh: bool = True,
+    ) -> None:
         if self.is_import:
             # skip updates for imported symbols
             # just bump the version if it's newly created
-            if self._timestamp == -1:
-                self._timestamp = nbs().cell_counter()
+            if not self._timestamp.is_initialized:
+                self._timestamp = Timestamp.current()
+            return
+        if mutated and self.obj_type in self.IMMUTABLE_TYPES:
             return
         # if we get here, no longer implicit
         self._implicit = False
@@ -479,10 +506,10 @@ class DataSymbol:
                 continue
             new_parent.children_by_cell_position[nbs().active_cell_position_idx].add(self)
             self.parents.add(new_parent)
-        self.required_timestamp = -1
+        self.required_timestamp = Timestamp.uninitialized()
         equal_to_old = self.prev_obj_definitely_equal_to_current_obj(prev_obj)
         if mutated or isinstance(self.stmt_node, ast.AugAssign):
-            self.timestamp_by_used_time[nbs().cell_counter()] = self.timestamp
+            self.update_usage_info()
         if refresh:
             self.refresh(
                 bump_version=not equal_to_old,
@@ -495,21 +522,36 @@ class DataSymbol:
             UpdateProtocol(self)(new_deps, mutated, propagate_to_namespace_descendents)
         self._refresh_cached_obj()
         nbs().updated_symbols.add(self)
+        if self.is_class:
+            # pop pending class defs and update obj ref
+            pending_class_ns = tracer().pending_class_namespaces.pop()
+            pending_class_ns.update_obj_ref(self.obj)
+
+    def update_usage_info(self, used_time: Optional[Timestamp] = None, exclude_ns: bool = False) -> None:
+        if used_time is None:
+            used_time = Timestamp.current()
+        if nbs().is_develop:
+            logger.info('sym `%s` used in cell %d last updated in cell %d', self, used_time.cell_num, self.timestamp)
+        if used_time not in self.timestamp_by_used_time and self.timestamp < used_time:
+            self.timestamp_by_used_time[used_time] = (
+                self.timestamp_excluding_ns_descendents if exclude_ns else self.timestamp
+            )
 
     def refresh(
         self,
         bump_version=True,
         refresh_descendent_namespaces=False,
         refresh_namespace_stale=True,
-        timestamp: Optional[int] = None,
+        timestamp: Optional[Timestamp] = None,
         seen: Set[DataSymbol] = None,
-    ):
+    ) -> None:
         self._temp_disable_warnings = False
         self.fresher_ancestors.clear()
         if bump_version:
-            self._timestamp = nbs().cell_counter() if timestamp is None else timestamp
+            self._timestamp = Timestamp.current() if timestamp is None else timestamp
             ns = self.containing_namespace
             if ns is not None:
+                # logger.error("bump version of %s due to %s (value %s)", ns.full_path, self.full_path, self.obj)
                 ns.max_descendent_timestamp = self._timestamp
             self.updated_timestamps.add(self._timestamp)
             self._version += 1
@@ -530,6 +572,7 @@ class DataSymbol:
                     # `test_external_object_update_propagates_to_stale_namespace_symbols()`
                     # in `test_multicell_precheck.py`
                     if not dsym.is_stale or refresh_namespace_stale:
+                        # logger.error("refresh %s due to %s (value %s) via namespace %s", dsym.full_path, self.full_path, self.obj, ns.full_path)
                         dsym.refresh(refresh_descendent_namespaces=True, timestamp=self._timestamp, seen=seen)
             if refresh_namespace_stale:
                 self.namespace_stale_symbols.clear()
