@@ -1,5 +1,4 @@
-# -*- coding: utf-8 -*-
+# -*- coding: future_annotations -*-
 from .ast_helper import FastAst as fast
-from .dot_dict import DotDict
 from .misc_utils import KeyDict
 from .mixins import CommonEqualityMixin
